/*  This file is part of the KDE project
    Copyright (C) 2000 Simon Hausmann <hausmann@kde.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "scheduler.h"

#include "jobclasses.h"
#include "slavebase.h"
#include "slave.h"

#include <dcopclient.h>
#include <dcopobject.h>
#include <kdebug.h>

#include <stdlib.h>
#include <assert.h>
#include <unistd.h>

using namespace KIO;

Scheduler *Scheduler::s_self = 0;

Scheduler::Scheduler()
{
    connect( &m_timer, SIGNAL( timeout() ),
             this, SLOT( slotStep() ) );
}

Scheduler::~Scheduler()
{
}

bool Scheduler::doJob( SimpleJob *job )
{
    KURL url = job->url();

    if ( !SlaveBase::knownProtocol( url.protocol() ) )
        return false;

    ProtocolInfo *nfo = m_schedule[ url.protocol() ];

    if ( !nfo )
    {
        nfo = new ProtocolInfo;
        m_schedule.insert( url.protocol(), nfo );
    }

    if ( nfo->m_jobs.findRef( job ) != -1 )
        return false;

    nfo->m_jobs.append( job );

    m_timer.start( 0, true );

    return true;
}

void Scheduler::releaseJob( SimpleJob *job, bool _killSlave )
{
    kdDebug() << "releaseJob " << job << endl;

    // the job's url might have changed or something, so let's iterate
    // over all entries
    QDictIterator<ProtocolInfo> it( m_schedule );
    for (; it.current(); ++it )
    {
        kdDebug() << it.currentKey() << " : current " << it.current()->m_jobs.current() << endl;
        if ( it.current()->m_jobs.first() == job )
        {
            it.current()->m_running = false;
            if ( _killSlave )
                killSlave( it.current() );
        }

        it.current()->m_jobs.removeRef( job );
    }

    m_timer.start( 0, true );
}

void Scheduler::slotStep()
{
    QDictIterator<ProtocolInfo> it( m_schedule );
    for (; it.current(); ++it )
        doStep( it.current() );

    it.toFirst();
    while ( it.current() )
    {
        if ( it.current()->m_jobs.count() == 0 &&
             it.current()->m_slave == 0 )
            delete m_schedule.take( it.currentKey() );
        else
            ++it;
    }
}

void Scheduler::doStep( ProtocolInfo *nfo )
{
    if ( nfo->m_running )
        return;

    if ( nfo->m_jobs.count() == 0 )
    {
        //work done.

        killSlave( nfo );

        return;
    }

    nfo->m_running = true;

    SimpleJob *job = nfo->m_jobs.first();

    if ( !nfo->m_slave )
    {
        kdDebug() << "creating slave for " << job->url().prettyURL() << endl;
        nfo->m_slave = createSlave( job->url().protocol() );
    }

    job->start( nfo->m_slave );
}

void Scheduler::killSlave( ProtocolInfo *nfo )
{
    if ( nfo->m_slave )
        nfo->m_slave->kill();

    delete nfo->m_slave;
    nfo->m_slave = 0;
}

Scheduler *Scheduler::self()
{
    if ( !s_self )
        s_self = new Scheduler;
    return s_self;
}

Slave *Scheduler::createSlave( const QString &prot )
{
    int app2slave[ 2 ]; // app -> slave
    int slave2app[ 2 ]; // slave -> app
    int dcopApp2Slave[ 2 ]; // dcop: app -> slave
    int dcopSlave2App[ 2 ]; // dcop: slave -> app

    if ( pipe( app2slave ) == -1 )
    {
        perror( "FATAL: can't create pipe for app->slave communication" );
        ::exit( 0 );
    }

    if ( pipe( slave2app ) == -1 )
    {
        perror( "FATAL: can't create pipe for slave->app communcation" );
        ::exit( 0 );
    }

    if ( pipe( dcopApp2Slave ) == -1 )
    {
        perror( "FATAL: can't create pipe for dcop app->slave communcation" );
        ::exit( 0 );
    }

    if ( pipe( dcopSlave2App ) == -1 )
    {
        perror( "FATAL: can't create pipe for dcop slave->app communication" );
        ::exit( 0 );
    }

    pid_t pid = fork();

    if ( pid == 0 )
    { // child


        // close unused fds
        ::close( app2slave[ 1 ] );
        ::close( slave2app[ 0 ] );

        ::close( dcopApp2Slave[ 1 ] );
        ::close( dcopSlave2App[ 0 ] );

        Connection *connection = new Connection;
        connection->init( app2slave[ 0 ], slave2app[ 1 ] );

        SlaveBase *slave = SlaveBase::createSlave( prot );

        assert( slave );

        slave->setConnection( connection );

        connection = new Connection;
        connection->init( dcopApp2Slave[ 0 ], dcopSlave2App[ 1 ] );

        DCOPClient::setGlobalConnection( connection );

        slave->dispatchLoop();

        // never reached

        ::exit( 0 );
    }

    assert( pid != -1 );

    // parent

    ::close( app2slave[ 0 ] );
    ::close( slave2app[ 1 ] );

    ::close( dcopApp2Slave[ 0 ] );
    ::close( dcopSlave2App[ 1 ] );

    Connection *connection = new Connection;
    connection->init( slave2app[ 0 ], app2slave[ 1 ] );

    Slave *res = new Slave( connection, pid );

    connection = new Connection;
    connection->init( dcopSlave2App[ 0 ], dcopApp2Slave[ 1 ] );

    DCOPObject::self()->addClient( connection, res );

    return res;
}

#include "scheduler.moc"
