/*  This file is part of the KDE project
    Copyright (C) 2000 Simon Hausmann <hausmann@kde.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "slavebase.h"
#include "job.h"

#include "httpfactory.h"
#include "httpsfactory.h"
#include "kio_file.h"

#include "slaveinterface.h"
#include "connection.h"

#include <qdatastream.h>
#include <qcstring.h>

#include <kdebug.h>

#include <assert.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>

using namespace KIO;

#define KIO_ARGS QByteArray data; QDataStream stream( data, IO_WriteOnly ); stream

SlaveBase *SlaveBase::createSlave( const QString &prot )
{
    SlaveBase * res = 0;

    if ( prot == "http" )
        res = createHTTPProtocolInstance();
    else if ( prot == "https" )
        res = createHTTPSProtocolInstance();
    else if ( prot == "file" )
        res = new File();

    return res;
}

bool SlaveBase::knownProtocol( const QString &protocol )
{
    return ( protocol == "http" || protocol == "https" || protocol == "file" );
}

SlaveBase::SlaveBase( const QCString &protocol, const QCString &, const QCString & )
{
    mProtocol = protocol;
    m_connection = 0;
}

SlaveBase::~SlaveBase()
{
    delete m_connection;
}

void SlaveBase::dispatchLoop()
{
    assert( m_connection );

    fd_set fds;
    struct timeval tv;
    int retval;

    while ( 42 )
    {
        FD_ZERO( &fds );
        FD_SET( m_connection->fd_from(), &fds );

        tv.tv_sec = 30;
        tv.tv_usec = 0;

        retval = select( m_connection->fd_from() + 1, &fds, NULL, NULL, &tv );

        if ( retval > 0 )
        {
            if ( FD_ISSET( m_connection->fd_from(), &fds ) )
            {
                int cmd;
                QByteArray data;
                if ( m_connection->read( &cmd, data ) != -1 )
                    dispatch( cmd, data );
                else
                {
                    kdDebug() << "SlaveBase: read() error! exiting!" << endl;
                    ::exit( 0 );
                }
            }
        }
        else if ( retval == -1 )
        {
            kdDebug() << "SlaveBase: select() error! exiting!" << endl;
            ::exit( 0 );
        }
    }
}

void SlaveBase::data( const QByteArray &data )
{
    sendMetaData();
    m_connection->send( MSG_DATA, data );
}

void SlaveBase::dataReq()
{
    sendMetaData();

    m_connection->send( MSG_DATA_REQ );
}

void SlaveBase::error( int _errid, const QString &_text )
{
    m_incomingMetaData.clear();
    m_outgoingMetaData.clear();

    KIO_ARGS << _errid << _text;

    m_connection->send( MSG_ERROR, data );
}

void SlaveBase::connected()
{
}

void SlaveBase::finished()
{
    sendMetaData();
    m_connection->send( MSG_FINISHED );
}

void SlaveBase::slaveStatus( const QString &, bool )
{
}

void SlaveBase::statEntry( const UDSEntry & )
{
}

void SlaveBase::totalSize( unsigned long )
{
}

void SlaveBase::processedSize( unsigned long )
{
}

void SlaveBase::speed( unsigned long )
{
}

void SlaveBase::redirection( const KURL &url )
{
    KIO_ARGS << url;
    m_connection->send( INF_REDIRECTION, data );
}

void SlaveBase::errorPage()
{
}

void SlaveBase::mimeType( const QString & )
{
}

void SlaveBase::infoMessage( const QString &msg )
{
    KIO_ARGS << msg;
    m_connection->send( INF_INFOMESSAGE, data );
}

bool SlaveBase::openPassDlg( AuthInfo &info )
{
    KIO_ARGS << info.prompt << info.username << info.caption
             << info.comment << info.commentLabel << info.readOnly;

    int cmd;

    m_connection->send( INF_NEED_PASSWD, data );

    QByteArray replyData;

    if ( waitForAnswer( CMD_USERPASS, CMD_NONE, &cmd, replyData ) != -1 &&
         cmd == CMD_USERPASS )
    {
        QDataStream replyStream( replyData, IO_ReadOnly );
        replyStream >> info.username >> info.password;
        return true;
    }

    return false;
}

int SlaveBase::messageBox( int type, const QString &text,
                const QString &caption,
                const QString &buttonYes,
                const QString &buttonNo )
{
    KIO_ARGS << type << text << caption << buttonYes << buttonNo;

    m_connection->send( INF_MESSAGEBOX, data );

    QByteArray replyData;

    if ( waitForAnswer( CMD_MESSAGEBOXANSWER, 0, 0, replyData ) != -1 )
    {
        QDataStream replyStream( replyData, IO_ReadOnly );
        int answer;

        replyStream >> answer;

        return answer;
    }
    else
    {
        assert( false );
        return 0;
    }
}

void SlaveBase::setMetaData( const QString &key, const QString &value )
{
    m_outgoingMetaData.insert( key, value );
}

QString SlaveBase::metaData( const QString &key )
{
    MetaData::ConstIterator it = m_incomingMetaData.find( key );

    if ( it == m_incomingMetaData.end() )
        return QString::null;

    return it.data();
}

void SlaveBase::setHost( const QString &, int , const QString &, const QString & )
{
}

void SlaveBase::get( const KURL & )
{
}

void SlaveBase::put( const KURL &, int , bool , bool )
{
}

void SlaveBase::stat( const KURL & )
{
}

void SlaveBase::mimetype( const KURL & )
{
}

void SlaveBase::special( const QByteArray & )
{
}

void SlaveBase::slave_status()
{
}

void SlaveBase::reparseConfiguration()
{
}

int SlaveBase::readData( QByteArray &buffer )
{
    return waitForAnswer( MSG_DATA, 0, 0, buffer );
}

bool SlaveBase::checkCachedAuthentication( AuthInfo &info )
{
    return false;
}

void SlaveBase::cacheAuthentication( const KURL &url, const QString &user, const QString &passwd,
                                     const QString &realm, const QString &extra )
{
}

void SlaveBase::sendMetaData()
{
    if ( m_outgoingMetaData.isEmpty() )
        return;

    KIO_ARGS << m_outgoingMetaData;
    m_connection->send( INF_META_DATA, data );
    m_outgoingMetaData.clear();
}

void SlaveBase::dispatch( int command, const QByteArray &packedArgs )
{
    QDataStream stream( packedArgs, IO_ReadOnly );

    KURL url;

    switch ( command )
    {
        case CMD_GET:
            stream >> url;
            get( url );
            break;
        case CMD_META_DATA:
            stream >> m_incomingMetaData;
            break;
        case CMD_SPECIAL:
            special( packedArgs );
            break;
        case CMD_HOST:
        {
            QString host, user, pass;
            Q_INT32 port;
            stream >> host >> port >> user >> pass;
            setHost( host, port, user, pass );
        }
        break;
        default:
            assert( 0 );
    }
}

int SlaveBase::waitForAnswer( int expected1, int expected2, int *_cmd, QByteArray &data )
{
    int res;
    int cmd;

    while ( 42 )
    {
        res = m_connection->read( &cmd, data );

        if ( res == -1 )
            return res;

        if ( cmd == expected1 || cmd == expected2 )
        {
            if ( _cmd )
                *_cmd = cmd;
            return res;
        }

        // ###
        if ( cmd == CMD_META_DATA )
            dispatch( cmd, data );
        else
            kdDebug() << "received unexpected command " << cmd << " while actually waiting for " << expected1 << endl;
    }
}

