#ifndef __kio_slavebase_h__
#define __kio_slavebase_h__

#include <qcstring.h>
#include <qstring.h>

#include <kio/global.h>
#include <kurl.h>

namespace KIO
{
/**
 * Structure to hold all authorization information.
 *
 * @param username          name as supplied and/or specified by user.
 *
 * @param password          password specified by user.
 *
 * @param prompt            information requested from user.
 *
 * @param caption           text displayed in the title bar of password dialog.
 *
 * @param comment           optional field for additional comment to relay to user.
 *
 * @param commentLabel      optional label to for additional comment (ex:"Command:").
 *
 * @param realmValue        optional field to specify "REALM" value for protocols that
 *                          use/support digest authentication.
 *
 * @param digestInfo        optional field to store any extra authorization string.  This
 *                          parameter is also only useful for digest authentication.
 * @param verifyPath
 *
 * @param readOnly          flag that forces the username field to be read-only if true.
 *
 * @param keepPassword      flag that forces the password to be cached for the duration of
 *                          the KDE session instead of it being automatically removed when
 *                          the application is terminated.
 */
struct AuthInfo
{
    AuthInfo() {
        verifyPath = false;
        readOnly = false;
        keepPassword = false;
    }

    KURL url;
    QString username;
    QString password;
    QString prompt;
    QString caption;
    QString comment;
    QString commentLabel;
    QString realmValue;
    QString digestInfo;
    bool verifyPath;
    bool readOnly;
    bool keepPassword;
};

    class Connection;

    class SlaveBase
    {
    public:
        static SlaveBase *createSlave( const QString &protocol );
        // ### hack until I implement the factories :)
        static bool knownProtocol( const QString &protocol );

        SlaveBase( const QCString &protocol, const QCString &pool_socket = QCString(), const QCString &app_socket = QCString() );
        virtual ~SlaveBase();

        void setConnection( Connection *conn ) { m_connection = conn; }
        Connection *connection() const { return m_connection; }

        void dispatchLoop();

        void data( const QByteArray &data );

        void dataReq();

        void error( int _errid, const QString &_text );

        void connected();

        void finished();

        void slaveStatus( const QString &host, bool connected );

        void statEntry( const UDSEntry &entry );

        void totalSize( unsigned long bytes );

        void processedSize( unsigned long bytes );

        void speed( unsigned long bytesPerSecond );

        void redirection( const KURL &url );

        void errorPage();

        void mimeType( const QString &mimeType );

        void infoMessage( const QString &msg );

        bool openPassDlg( AuthInfo &info );

        enum { QuestionYesNo = 1, WarningYesNo = 2, WarningContinueCancel = 3, WarningYesNoCancel = 4,
               Information = 5, SSLMessageBox = 6 };

        int messageBox( int type, const QString &text,
                        const QString &caption = QString::null,
                        const QString &buttonYes = QString::null,
                        const QString &buttonNo = QString::null );

        void setMetaData( const QString &key, const QString &value );

        QString metaData( const QString &key );

        virtual void setHost( const QString &host, int port, const QString &user, const QString &pass );

        virtual void get( const KURL &url );

        virtual void put( const KURL &url, int permissions, bool overwrite, bool resume );

        virtual void stat( const KURL &url );

        virtual void mimetype( const KURL &url );

        virtual void special( const QByteArray & );

        virtual void slave_status();

        virtual void reparseConfiguration();

        int readData( QByteArray &buffer );

        bool checkCachedAuthentication( AuthInfo &info );

        void cacheAuthentication( const KURL &url, const QString &user, const QString &passwd,
                                  const QString &realm = QString::null, const QString &extra = QString::null );

        void sendMetaData();

        void dispatch( int command, const QByteArray &packedArgs );

        void canResume() {}

    protected:
        QCString mProtocol;

    private:
        int waitForAnswer( int expected1, int expected2, int *_cmd, QByteArray &data );

        Connection *m_connection;
        MetaData m_incomingMetaData;
        MetaData m_outgoingMetaData;
    };

};

#endif
